import glob
import math
import os
import traceback

import tqdm

import gxl_ai_utils.AiConstant
from gxl_ai_utils.utils import utils_file, utils_spider

logger = gxl_ai_utils.AiConstant.AI_LOGGER()

PAGE_SIZE = 30
SORT_TYPE = 0  # 0: 正序,1:倒序
SAVE_WITH_TITLE = True
SAVE_WITH_ALBUM_ID = True


class GxlSpider(object):
    def __init__(self, output_root='./output/ximalaya/'):
        self.root_url = "https://www.ximalaya.com/category"
        self.root_start_url = "https://www.ximalaya.com/category"
        self.output_root = output_root
        utils_file.makedir_sil(self.output_root)

    def get_category_info(self):
        """
        得到每个类别的地址，存储为output_root/category.json
        :return:
        """
        xpath_href = '//*[@id="award"]/main/div[1]/div[2]/div/a/@href'
        xpath_title = '//*[@id="award"]/main/div[1]/div[2]/div/a/text()'
        hrefs = utils_spider.hande_href(self.root_start_url, xpath_href)
        titles = utils_spider.hande_href(self.root_start_url, xpath_title)
        datas = {}
        for title, href in zip(titles, hrefs):
            datas[title] = utils_file.join_path(self.root_start_url, href.split('/')[2])
        utils_file.print_dict(datas)
        utils_file.write_dict_to_json(datas, utils_file.join_path(self.output_root, 'category.json'))
        # threads = []
        # for item in datas.items():
        #     if item[0] == "音乐":
        #         logger.info('不再处理音乐专辑音乐')
        #         continue
        #     thread = threading.Thread(target=handle_category_index_fenye, args=(item[0], item[1]))
        #     thread.start()
        #     threads.append(thread)
        # for thread in threads:
        #     thread.join()

    def handle_category_fenye(self):
        """
        分页处理函数
        通过category.json文件数据， 分别获取每个类别的所有专辑的ID list, 并存入文件保存。
        文件保存地址为 output_root/album_info/{category}.list
        """
        category_info = utils_file.load_dict_from_json(os.path.join(self.output_root, 'category.json'))
        for category, url in category_info.items():
            logger.info(f'开始处理{category}')
            utils_file.makedir_for_file_or_dir(os.path.join(self.output_root, 'album_info', f'{category}.list'))
            utils_file.remove_file(os.path.join(self.output_root, 'album_info', f'{category}.list', 'index'))
            # 先得到该类别的分页数
            xpath = '//*[@id="award"]/main/div[1]/div[3]/div[3]/nav/div/form/input[2]/@max'
            res = utils_spider.hande_href(url, xpath)
            num_fenye = int(res[0])
            logger.info(f'该类别的分页数为{num_fenye}')
            try:
                self.handle_category_index(category, url)
            except Exception as e:
                logger.info(f'分页1的handle_category_index函数执行报错, 爆错内容为{e}')
                traceback.print_exc()
            for i in range(2, num_fenye + 1):
                try:
                    self.handle_category_index(category, utils_file.join_path(url, f'p{i}'))
                except Exception as e:
                    logger.info(f'分页{i}的handle_category_index函数执行报错, 爆错内容为{e}')
                    traceback.print_exc()
            break

    def handle_category_index(self, category, url):
        """
        得到当前类别的当前分页的所有的album ID, 并写入文件保存。
        :param category:
        :param url:
        :return:
        """
        logger.info(f'开始加载 {category} 类别的如下分页页面:' + url + ' ------------------------------')
        logger.info(f'加载成功, 开始处理 {category} 主页')
        xpath = '//ul[@class="_ZV"]/li[@class="_ZV"]/div[@class="album-wrapper   undefined T_G"]//div[@class="album-wrapper-card T_G"]/a[@class="album-cover   lg needhover _hW"]'
        datas = utils_spider.hande_href(url, xpath)
        xpath_for_title = '//*[@id="award"]/main/div[1]/div[3]/div[2]/ul/li/div/a[1]/span/text()'
        titles = utils_spider.hande_href(url, xpath_for_title)
        res = []
        res2 = []
        res3 = []
        titles_res = []
        for data in datas:
            res.append(data.xpath(".//@href"))
            res2.append(data.xpath(".//img[@class='corner-img _hW']/@src"))  # 付费专辑特有图标
            for re1, re2, title in zip(res, res2, titles):
                if len(re2) == 0:
                    res3.append(re1)
                    titles_res.append(title)
        res3 = [x[0] for x in res3]
        res3 = set(res3)
        logger.info(f"处理{category}主页成功, 得到当前分页的每个免费专辑的地址:")  # 删除付费专辑
        utils_file.print_list(list(res3))
        res3 = [x.split('/')[-1] for x in res3]
        utils_file.write_list_to_file(res3,
                                      utils_file.join_path(self.output_root, 'album_info', f'{category}.list'), True)

    def handle_album_fenye(self):
        """
        通过album_info 目录，得到每个分类下的每个专辑下的所有音频链接。并存储到文件中。

        :return:
        """
        albums_path_list = glob.glob(os.path.join(self.output_root, 'album_info', '*.list'))
        utils_file.print_list(albums_path_list)
        for albums_path in albums_path_list:
            category = os.path.basename(albums_path).split('.')[0]
            logger.info(f'开始得到{category}专辑的音频链接')
            utils_file.makedir_sil(os.path.join(self.output_root, 'audio_info', category))
            album_id_list = utils_file.load_list_file_clean(albums_path)
            for album_id in tqdm.tqdm(album_id_list):
                self.handle_album(category, album_id)

    def handle_album(self, category, album_id):
        """
        获得每个专辑的所有音频地址，并存储为文件， 文件位置为：
        output_root/audio_info/{category}/{album_id}.list
        album_id:型如 '9723091'
        分页获取数据
        audio:
        https://aod.cos.tx.xmcdn.com/storages/3842-audiofreehighqps/AC/76/GKwRIasJOqrJAU25xAKDendP.m4a
        https://aod.cos.tx.xmcdn.com/storages/eb2c-audiofreehighqps/48/79/GKwRIUEJDNFhARzq2gJyonzx.m4a


        """
        url_data = f"https://www.ximalaya.com/revision/album/v1/getTracksList?albumId={str(album_id)}&pageNum={1}&sort={SORT_TYPE}&pageSize={PAGE_SIZE}"
        logger.info(f'开始处理专辑(伴随分页模式),album_id:{album_id} ========')
        logger.info('开始获取总页数')
        response = utils_spider.send_request(url_data)
        if response is None:
            return
        json_data = response.json()
        if json_data is None:
            return
        try:
            total_num = json_data['gxl_data']['trackTotalCount']
        except Exception as e:
            logger.info('json_data解析失败,未在指定路径获取该专辑音频总数')
            return
        # total_page = int(total_num / 30) + 1
        total_page = math.ceil(int(total_num) / 30)
        logger.info(f'总页数为:{total_page}')
        utils_file.remove_file(os.path.join(self.output_root, 'audio_info', category, f'{album_id}.list'))
        for i in range(1, total_page + 1):
            logger.info(f'开始处理专辑{album_id}的第{i}页')
            url = f"https://www.ximalaya.com/revision/album/v1/getTracksList?albumId={album_id}&pageNum={i}&sort={SORT_TYPE}&pageSize={PAGE_SIZE}"
            response = utils_spider.send_request(url)
            if response is None:
                continue
            json_data = response.json()
            if json_data is None:
                continue
            try:
                data_list = json_data['gxl_data']['tracks']
                ids_list = [track['trackId'] for track in data_list]
                title_list = [track['title'] for track in data_list]
            except Exception as e:
                logger.info('json_data解析失败,未在指定路径获取该专辑音频id列表')
                continue
            ids = ids_list
            logger.info(f'处理专辑成功,得到该专辑所有音频id为:{ids}')
            logger.info(f'处理专辑成功,得到该专辑所有音频titles为:{title_list}')
            if title_list is None:
                title_list = []
            logger.info(f'开始利用异步请求从ids得到每个id对应的音频路径')
            if len(ids) != len(title_list):
                logger.info("ids与titles长度不一致, title并不重要, 使其强行一致即可")
                if len(ids) > len(title_list):
                    for _ in range(len(ids) - len(title_list)):
                        title_list.append("")
                else:
                    title_list = title_list[:len(ids)]
            res = []
            for id, title in zip(ids, title_list):
                logger.info(f'开始加载音频{id}的json数据')
                url = f"https://www.ximalaya.com/revision/play/v1/audio?id={id}&ptype=1"
                response = utils_spider.send_request(url)
                if response is None:
                    continue

                json_res = response.json()
                if json_data is None:
                    continue
                else:
                    logger.info(f'音频{id}的json数据加载成功')

                try:
                    src = json_res['gxl_data']['src']
                except Exception as e:
                    logger.info(f'音频{id}的json数据解析失败,未在指定路径得到音频地址')
                    continue
                logger.info(f'音频{id}的json数据解析成功,开始保存音频')
                res.append({'title': title, 'src': src, "id": id})
            utils_file.print_list(res)
            utils_file.write_dict_list_to_jsonl(res, os.path.join(self.output_root, 'audio_info', category,
                                                                 f'{album_id}.list'), True)


if __name__ == '__main__':
    spider = GxlSpider()
    # spider.get_category_info()
    # spider.handle_category_fenye()
    spider.handle_album_fenye()



import os

from gxl_ai_utils.audio_dataset import common_data_handler

from gxl_ai_utils.utils import utils_file, utils_data
from gxl_ai_utils.AiConstant import AI_LOGGER

def make_shard():
    source_dir = '/home/work_nfs7/xlgeng/gxl_data/asr_data_inventory/'
    output_dir_source = '/home/backup_nfs5/xlgeng/gxl_data/shard_asr/'
    utils_file.makedir_sil(output_dir_source)
    logger = AI_LOGGER('./output/make_shard_to_data_big.log')
    for root, dirs, files in os.walk(source_dir):
        for dir in dirs:
            logger.info(f'开始处理{dir}目录')
            wav_scp = utils_file.join_path(root, dir, "wav.scp")
            text_scp = utils_file.join_path(root, dir, "text")
            if not os.path.exists(wav_scp) or not os.path.exists(text_scp):
                logger.info(f'skip {dir}, the dir do not have wav.scp or text file')
                continue
            if dir == "gigaspeech":
                logger.info(f'skip {dir}, the dir is gigaspeech, and the dir has handled before')
                continue
            output_dir = utils_file.join_path(output_dir_source, dir)
            # size = utils_file.get_dir_size(output_dir)
            # if size > 10:
            #     logger.info(f'skip {dir}, size:{size} the dir is big, and the dir has handled before')
            #     continue
            # if os.path.exists(output_dir):
            #     shutil.rmtree(output_dir)
            common_data_handler.make_shard_file(wav_scp, text_scp, output_dir, num_utt_per_shard=1000,
                                                prefix_for_tar_file=dir, logger=logger)
        break

if __name__ == '__main__':
    make_shard()



import math

from matplotlib import pyplot as plt
import matplotlib.pyplot as plt
import numpy as np
import matplotlib

# 设置matplotlib正常显示中文和负号
matplotlib.rcParams['font.sans-serif'] = ['SimHei']  # 用黑体显示中文
matplotlib.rcParams['axes.unicode_minus'] = False  # 正常显示负号


def show_image_for_arrays(image_arrays, text_labels, col_num=5):
    """展示数组形式的图片"""
    num = len(image_arrays)
    row_num = math.ceil(num / col_num)
    plt.figure(figsize=(20, 6 * row_num))
    for i in range(1, num + 1):
        plt.subplot(row_num, col_num, i)
        plt.imshow(image_arrays[i - 1])
        plt.xticks([])
        plt.yticks([])
        plt.title(text_labels[i - 1])
    plt.subplots_adjust(hspace=0, wspace=0)  # 调整子图间距
    plt.show()


def show_hist(data, bins=50):
    """
    绘制直方图
    gxl_data:必选参数，绘图数据
    bins:直方图的长条形数目，可选项，默认为10
    facecolor:长条形的颜色
    edgecolor:长条形边框的颜色
    alpha:透明度
    """
    plt.hist(data, bins=bins, facecolor="blue", edgecolor="black", alpha=0.7)
    # 显示横轴标签
    plt.xlabel("区间")
    # 显示纵轴标签
    plt.ylabel("频数/频率")
    # 显示图标题
    plt.title("频数/频率分布直方图")
    plt.show()


def show_lines(x, y):
    plt.plot(x, y)
    plt.grid(True, linestyle='--', linewidth=0.5, color='gray', alpha=0.5)
    plt.show()


def show_func(func: callable, x_min=-3, x_max=3):
    x = np.arange(x_min, x_max, 0.01)
    y = func(x)
    show_lines(x, y)



# from distutils.core import setup
from setuptools import setup, find_packages

# packages = ['gxl_ai_utils', 'gxl_ai_utils.utils',
#             'gxl_ai_utils.config', 'gxl_ai_utils.run',
#             'gxl_ai_utils.thread' ]
setup(name='gxl_ai_utils',
      version='1.1.5',
      author='Xuelong Geng',
      description='这个是耿雪龙的工具包模块. update time: 2024-1-16',
      author_email='3349495429@qq.com',
      packages=find_packages(),
      package_dir={'requests': 'requests'}, )



from matplotlib import pyplot as plt

from gxl_ai_utils.utils import utils_file


def main():
    """
    统计喜马拉雅音频下载的数量，并绘制成柱状图
    :return:
    """
    res_dic = {}
    with open('gxl.txt', 'r', encoding='utf-8') as f:
        for line in f:
            print(line)
            items = line.strip().split(' ')
            res_dic[items[-1].strip()] = int(items[0].strip())
    with open('gxl2.txt', 'r', encoding='utf-8') as f:
        for line in f:
            print(line)
            items = line.strip().split(' ')
            if items[-1].strip() not in res_dic:
                res_dic[items[-1].strip()] = int(items[0].strip())
            else:
                res_dic[items[-1].strip()] = int(res_dic[items[-1].strip()]) + int(items[0].strip())
    print(res_dic)
    utils_file.write_dict_to_json(res_dic, './res.json')
    items_list = list(res_dic.items())
    items_list.sort(key=lambda x: x[1])
    print(items_list)
    x_labels = [x for x, y in items_list]
    y_labels = [y for x, y in items_list]
    print(x_labels)
    new_labels = []
    for l in x_labels:
        if len(l) > 2:
            l = l[:2] + '\n' + l[2:]
        new_labels.append(l)
    print(y_labels)
    # x = ['A', 'B', 'C', 'D', 'E']  # x轴的标签
    # y = [10, 20, 15, 25, 30]  # y轴的数值
    total = 0
    for y in y_labels:
        total += y
    # 创建一个图形对象
    fig = plt.figure()

    # 在图形对象上添加一个子图
    ax = fig.add_subplot(111)

    # 在子图上绘制柱状图
    rects = ax.bar(new_labels, y_labels, color='blue', width=0.8)

    # 在每个柱子的顶部添加数值标签
    for rect in rects:
        height = rect.get_height()  # 获取柱子的高度
        x_pos = rect.get_x() + rect.get_width() / 2  # 获取柱子的中心位置
        y_pos = height + 0.5  # 获取文本标签的纵坐标位置，偏移量为0.5
        label = f'{height:.0f}'  # 获取文本标签的内容，保留一位小数
        ax.text(x_pos, y_pos, label, ha='center', va='bottom', fontsize=12)  # 添加文本标签

    # 设置x轴和y轴的标签
    ax.set_xlabel('Category')
    ax.set_ylabel('Value/GB')

    # 设置图形的标题
    ax.set_title(f'每个目录的大小，总大小：{total}GB')

    # 保存图形为图片，格式为png，分辨率为300dpi
    fig.savefig('bar_chart.png', format='png', dpi=600)


if __name__ == '__main__':
    main()



import threading
import multiprocessing
from functools import partial


class GxlDynamicThreadPool:
    def __init__(self, ):
        self._threads = []

    def add_thread(self, func: callable, fun_args: list):
        thread = threading.Thread(target=func, args=fun_args)
        self._threads.append(thread)

    def add_task(self, func: callable, fun_args: list):
        self.add_thread(func, fun_args)

    def start(self):
        for thread in self._threads:
            thread.start()
        self._join()

    def _join(self):
        for thread in self._threads:
            thread.join()


class GxlFixedThreadPool:
    def __init__(self, num_threads: int):
        self.pool = multiprocessing.Pool(processes=num_threads)

    def map(self, func: callable,iterable_arg_list:list, other_fun_args: dict):
        """
        要求函数只用第一个参数是遍历list。
        :param func:
        :param fun_args:
        :return:
        """
        partial_my_function = partial(func, **other_fun_args)
        return self.pool.map(partial_my_function, iterable_arg_list)


    def apply_async(self, func: callable, fun_args: list):
        self.pool.apply_async(func, fun_args)

    def add_thread(self, func: callable, fun_args: list):
        self.apply_async(func, fun_args)

    def add_task(self, func: callable, fun_args: list):
        self.apply_async(func, fun_args)

    def start(self):
        self.pool.close()
        self.pool.join()



# Copyright (c) 2021 Mobvoi Inc. (authors: Binbin Zhang, Di Wu)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Modified from ESPnet(https://github.com/espnet/espnet)
"""Decoder definition."""
from typing import Tuple, List, Optional

import torch

from ..transformer.attention import MultiHeadedAttention
from ..transformer.decoder_layer import DecoderLayer
from ..transformer.embedding import PositionalEncoding
from ..transformer.embedding import NoPositionalEncoding
from ..transformer.positionwise_feed_forward import PositionwiseFeedForward
from ..utils.mask import (subsequent_mask, make_pad_mask)


class TransformerDecoder(torch.nn.Module):
    """Base class of Transfomer decoder module.
    Args:
        vocab_size: output dim
        encoder_output_size: dimension of attention
        attention_heads: the number of heads of multi head attention
        linear_units: the hidden units number of position-wise feedforward
        num_blocks: the number of decoder blocks
        dropout_rate: dropout rate
        self_attention_dropout_rate: dropout rate for attention
        input_layer: input layer type
        use_output_layer: whether to use output layer
        pos_enc_class: PositionalEncoding or ScaledPositionalEncoding
        normalize_before:
            True: use layer_norm before each sub-block of a layer.
            False: use layer_norm after each sub-block of a layer.
        src_attention: if false, encoder-decoder cross attention is not
                       applied, such as CIF model
    """

    def __init__(
        self,
        vocab_size: int,
        encoder_output_size: int,
        attention_heads: int = 4,
        linear_units: int = 2048,
        num_blocks: int = 6,
        dropout_rate: float = 0.1,
        positional_dropout_rate: float = 0.1,
        self_attention_dropout_rate: float = 0.0,
        src_attention_dropout_rate: float = 0.0,
        input_layer: str = "embed",
        use_output_layer: bool = True,
        normalize_before: bool = True,
        src_attention: bool = True,
    ):
        super().__init__()
        attention_dim = encoder_output_size

        if input_layer == "embed":
            self.embed = torch.nn.Sequential(
                torch.nn.Embedding(vocab_size, attention_dim),
                PositionalEncoding(attention_dim, positional_dropout_rate),
            )
        elif input_layer == 'none':
            self.embed = NoPositionalEncoding(attention_dim,
                                              positional_dropout_rate)
        else:
            raise ValueError(f"only 'embed' is supported: {input_layer}")

        self.normalize_before = normalize_before
        self.after_norm = torch.nn.LayerNorm(attention_dim, eps=1e-5)
        self.use_output_layer = use_output_layer
        self.output_layer = torch.nn.Linear(attention_dim, vocab_size)
        self.num_blocks = num_blocks
        self.decoders = torch.nn.ModuleList([
            DecoderLayer(
                attention_dim,
                MultiHeadedAttention(attention_heads, attention_dim,
                                     self_attention_dropout_rate),
                MultiHeadedAttention(attention_heads, attention_dim,
                                     src_attention_dropout_rate)
                if src_attention else None,
                PositionwiseFeedForward(attention_dim, linear_units,
                                        dropout_rate),
                dropout_rate,
                normalize_before,
            ) for _ in range(self.num_blocks)
        ])

    def forward(
        self,
        memory: torch.Tensor,
        memory_mask: torch.Tensor,
        ys_in_pad: torch.Tensor,
        ys_in_lens: torch.Tensor,
        r_ys_in_pad: torch.Tensor = torch.empty(0),
        reverse_weight: float = 0.0,
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """Forward decoder.
        Args:
            memory: encoded memory, float32  (batch, maxlen_in, feat)
            memory_mask: encoder memory mask, (batch, 1, maxlen_in)
            ys_in_pad: padded input token ids, int64 (batch, maxlen_out)
            ys_in_lens: input lengths of this batch (batch)
            r_ys_in_pad: not used in transformer decoder, in order to unify api
                with bidirectional decoder
            reverse_weight: not used in transformer decoder, in order to unify
                api with bidirectional decode
        Returns:
            (tuple): tuple containing:
                x: decoded token score before softmax (batch, maxlen_out,
                    vocab_size) if use_output_layer is True,
                torch.tensor(0.0), in order to unify api with bidirectional decoder
                olens: (batch, )
        """
        tgt = ys_in_pad
        maxlen = tgt.size(1)
        # tgt_mask: (B, 1, L)
        tgt_mask = ~make_pad_mask(ys_in_lens, maxlen).unsqueeze(1)
        tgt_mask = tgt_mask.to(tgt.device)
        # m: (1, L, L)
        m = subsequent_mask(tgt_mask.size(-1),
                            device=tgt_mask.device).unsqueeze(0)
        # tgt_mask: (B, L, L)
        tgt_mask = tgt_mask & m
        x, _ = self.embed(tgt)
        for layer in self.decoders:
            x, tgt_mask, memory, memory_mask = layer(x, tgt_mask, memory,
                                                     memory_mask)
        if self.normalize_before:
            x = self.after_norm(x)
        if self.use_output_layer:
            x = self.output_layer(x)
        olens = tgt_mask.sum(1)
        return x, torch.tensor(0.0), olens

    def forward_one_step(
        self,
        memory: torch.Tensor,
        memory_mask: torch.Tensor,
        tgt: torch.Tensor,
        tgt_mask: torch.Tensor,
        cache: Optional[List[torch.Tensor]] = None,
    ) -> Tuple[torch.Tensor, List[torch.Tensor]]:
        """Forward one step.
            This is only used for decoding.
        Args:
            memory: encoded memory, float32  (batch, maxlen_in, feat)
            memory_mask: encoded memory mask, (batch, 1, maxlen_in)
            tgt: input token ids, int64 (batch, maxlen_out)
            tgt_mask: input token mask,  (batch, maxlen_out)
                      dtype=torch.uint8 in PyTorch 1.2-
                      dtype=torch.bool in PyTorch 1.2+ (include 1.2)
            cache: cached output list of (batch, max_time_out-1, size)
        Returns:
            y, cache: NN output value and cache per `self.decoders`.
            y.shape` is (batch, maxlen_out, token)
        """
        x, _ = self.embed(tgt)
        new_cache = []
        for i, decoder in enumerate(self.decoders):
            if cache is None:
                c = None
            else:
                c = cache[i]
            x, tgt_mask, memory, memory_mask = decoder(x,
                                                       tgt_mask,
                                                       memory,
                                                       memory_mask,
                                                       cache=c)
            new_cache.append(x)
        if self.normalize_before:
            y = self.after_norm(x[:, -1])
        else:
            y = x[:, -1]
        if self.use_output_layer:
            y = torch.log_softmax(self.output_layer(y), dim=-1)
        return y, new_cache


class BiTransformerDecoder(torch.nn.Module):
    """Base class of Transfomer decoder module.
    Args:
        vocab_size: output dim
        encoder_output_size: dimension of attention
        attention_heads: the number of heads of multi head attention
        linear_units: the hidden units number of position-wise feedforward
        num_blocks: the number of decoder blocks
        r_num_blocks: the number of right to left decoder blocks
        dropout_rate: dropout rate
        self_attention_dropout_rate: dropout rate for attention
        input_layer: input layer type
        use_output_layer: whether to use output layer
        pos_enc_class: PositionalEncoding or ScaledPositionalEncoding
        normalize_before:
            True: use layer_norm before each sub-block of a layer.
            False: use layer_norm after each sub-block of a layer.
    """

    def __init__(
        self,
        vocab_size: int,
        encoder_output_size: int,
        attention_heads: int = 4,
        linear_units: int = 2048,
        num_blocks: int = 6,
        r_num_blocks: int = 0,
        dropout_rate: float = 0.1,
        positional_dropout_rate: float = 0.1,
        self_attention_dropout_rate: float = 0.0,
        src_attention_dropout_rate: float = 0.0,
        input_layer: str = "embed",
        use_output_layer: bool = True,
        normalize_before: bool = True,
    ):

        super().__init__()
        self.left_decoder = TransformerDecoder(
            vocab_size, encoder_output_size, attention_heads, linear_units,
            num_blocks, dropout_rate, positional_dropout_rate,
            self_attention_dropout_rate, src_attention_dropout_rate,
            input_layer, use_output_layer, normalize_before)

        self.right_decoder = TransformerDecoder(
            vocab_size, encoder_output_size, attention_heads, linear_units,
            r_num_blocks, dropout_rate, positional_dropout_rate,
            self_attention_dropout_rate, src_attention_dropout_rate,
            input_layer, use_output_layer, normalize_before)

    def forward(
        self,
        memory: torch.Tensor,
        memory_mask: torch.Tensor,
        ys_in_pad: torch.Tensor,
        ys_in_lens: torch.Tensor,
        r_ys_in_pad: torch.Tensor,
        reverse_weight: float = 0.0,
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """Forward decoder.
        Args:
            memory: encoded memory, float32  (batch, maxlen_in, feat)
            memory_mask: encoder memory mask, (batch, 1, maxlen_in)
            ys_in_pad: padded input token ids, int64 (batch, maxlen_out)
            ys_in_lens: input lengths of this batch (batch)
            r_ys_in_pad: padded input token ids, int64 (batch, maxlen_out),
                used for right to left decoder
            reverse_weight: used for right to left decoder
        Returns:
            (tuple): tuple containing:
                x: decoded token score before softmax (batch, maxlen_out,
                    vocab_size) if use_output_layer is True,
                r_x: x: decoded token score (right to left decoder)
                    before softmax (batch, maxlen_out, vocab_size)
                    if use_output_layer is True,
                olens: (batch, )
        """
        l_x, _, olens = self.left_decoder(memory, memory_mask, ys_in_pad,
                                          ys_in_lens)
        r_x = torch.tensor(0.0)
        if reverse_weight > 0.0:
            r_x, _, olens = self.right_decoder(memory, memory_mask, r_ys_in_pad,
                                               ys_in_lens)
        return l_x, r_x, olens

    def forward_one_step(
        self,
        memory: torch.Tensor,
        memory_mask: torch.Tensor,
        tgt: torch.Tensor,
        tgt_mask: torch.Tensor,
        cache: Optional[List[torch.Tensor]] = None,
    ) -> Tuple[torch.Tensor, List[torch.Tensor]]:
        """Forward one step.
            This is only used for decoding.
        Args:
            memory: encoded memory, float32  (batch, maxlen_in, feat)
            memory_mask: encoded memory mask, (batch, 1, maxlen_in)
            tgt: input token ids, int64 (batch, maxlen_out)
            tgt_mask: input token mask,  (batch, maxlen_out)
                      dtype=torch.uint8 in PyTorch 1.2-
                      dtype=torch.bool in PyTorch 1.2+ (include 1.2)
            cache: cached output list of (batch, max_time_out-1, size)
        Returns:
            y, cache: NN output value and cache per `self.decoders`.
            y.shape` is (batch, maxlen_out, token)
        """
        return self.left_decoder.forward_one_step(memory, memory_mask, tgt,
                                                  tgt_mask, cache)



from typing import Optional

import torch
from torch import nn


class AbsEncoder(nn.Module):
    def output_size(self):
        raise NotImplementedError

    def forward(self, xs_pad: torch.Tensor,
                ilens: torch.Tensor,
                pre_states: torch.tensor = None
                ) -> tuple[torch.Tensor, torch.Tensor, Optional[torch.Tensor]]:
        raise NotImplementedError
